package mimuw.mmf;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import mimuw.mmf.clusterers.ClusterTools;
import mimuw.mmf.clusterers.DistanceCounter;
import mimuw.mmf.clusterers.WeighableClustersComparator;
import mimuw.mmf.clusterers.hierarchical.HierarchicalCluster;
import mimuw.mmf.clusterers.hierarchical.HierarchicalClusterer;
import mimuw.mmf.clusterers.hierarchical.HierarchicalClusterer.THRESHOLD_TYPE;
import mimuw.mmf.config.ApplicationData;
import mimuw.mmf.config.ConfigData;
import mimuw.mmf.config.ConfigSAXParser;
import mimuw.mmf.config.JasparData;
import mimuw.mmf.finders.FileFormatException;
import mimuw.mmf.finders.MMFException;
import mimuw.mmf.finders.MMFRuntimeException;
import mimuw.mmf.finders.MotifFinder;
import mimuw.mmf.finders.WrongValueReturnedException;
import mimuw.mmf.finders.bioprospector.BioProspectorFinder;
import mimuw.mmf.finders.mdscan.MDscanFinder;
import mimuw.mmf.finders.meme.MEMEFinder;
import mimuw.mmf.finders.weeder.WeederFinder;
import mimuw.mmf.jaspar.JasparDataReader;
import mimuw.mmf.jaspar.JasparMotif;
import mimuw.mmf.motif.AbstractMotif;
import mimuw.mmf.motif.Motif;
import mimuw.mmf.motif.MotifOccurrence;
import mimuw.mmf.motif.external.ExternalMotif;
import mimuw.mmf.motif.external.ExternalMotifsReader;
import mimuw.mmf.util.AlignmentTools;
import mimuw.mmf.util.BioTools;
import mimuw.mmf.util.Constants;
import mimuw.mmf.util.Logger;
import mimuw.mmf.util.Pair;
import mimuw.mmf.util.QuickBioTools;
import mimuw.mmf.util.SimilarityFilter;
import mimuw.mmf.util.QuickBioTools.ColumnsDistributionComparator;
import mimuw.mmf.util.QuickBioTools.FittingOnSequenceComparator;

import org.biojava.bio.BioException;
import org.biojava.bio.dist.Distribution;
import org.biojava.bio.seq.DNATools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.symbol.SymbolList;
import org.biojavax.bio.seq.RichSequence;
import org.xml.sax.SAXParseException;

public class MMF {

	/* Available clusterings */
	public enum CLUSTERING_TYPE {HIERARCHICAL, KMEDOIDS};
	
	/* Available functions to count the distance between distributions */
	public enum DISTRIBUTION_DISTANCE_FUNCTION {DPQ, RELENT, PEARSON};
	
	/* Available ways of obtaining distribution from motifs 
	 * COLUMNS - compare motifs' PSSM 
	 * FITTING - compare arrays of probabilities describing how well the motif fits on
	 * 	each position of a DNA sequence created by concatenating all input sequences */
	public enum DISTRIBUTION_TYPE {COLUMNS, FITTING};

	
	
	/**
	 * Internal class to represent one motif-finding program and value specifying whether 
	 * this program should be used or no (@registered)
	 */
	private class MotifFinderEntry {

		private MotifFinder finder;
		private boolean registered;

		private MotifFinderEntry (MotifFinder f){
			finder = f;
			registered = false;
		}

		private void register () { registered = true; }
		private boolean isRegistered () {return registered; }
		private MotifFinder getFinder(){ return finder;}
	}



	private Map<String, MotifFinderEntry> finders= new HashMap<String, MotifFinderEntry>();
	private File tmpFastaFile;
	private double[] background;


	
	private MMF(){}

	

	/**
	 * Registers the motif finder agent f. Only after calling this method, that agent will be launched to find motifs
	 * @param f
	 */
	private void register(MotifFinder f){
		finders.put(f.getProgramIdentifier(), new MotifFinderEntry(f));
	}



	
	private void setParams(ConfigData config, Map<String, String> parametersStringsMap) 
		throws MMFException, WrongParameterException
	{
		for (String finderId: parametersStringsMap.keySet()){
			MotifFinderEntry mfe = finders.get(finderId);
			if (mfe == null)
				throw new MMFException("Unrecognized entry - '" + finderId + "' - as program name");
			
			MotifFinder finder = mfe.getFinder();
			
			Integer len = config.getFindersData().getMotifLength();
			Boolean both = config.getFindersData().getBothStrands();
			Integer rep = config.getFindersData().getMotifsReported();
			
			if(len != null)	finder.setMotifLength(len);
			if(both != null)finder.setSearchBothStrands(both);
			if(rep != null)	finder.setDesiredNumberOfMotifs(rep);
			
			finder.setParametersAsSingleString(parametersStringsMap.get(finderId));
			mfe.register();
		}
	}

	
	
	private void reportSuccess(
			HierarchicalCluster<AbstractMotif> c, 
			int clustersIndex,
			int numberOfFindersInvolved,
			String classOfExpectedJasparMotif)
	{
			int motifsInExpectedCLuster = ClusterTools.getNumberOfNonJasparMotifs(c);
			String findersToString = ClusterTools.getFindersNamesAsString(c);
			if (clustersIndex > 0)
				Logger.auto("%d unexpected clusters preceed the one correct.", clustersIndex);
			Logger.auto("%s clustered with %d other motifs returned by %d finders: %s\n", 
					classOfExpectedJasparMotif,
					motifsInExpectedCLuster,
					numberOfFindersInvolved, 
					findersToString);
	}
	
	

	private void doClustering(
			List<AbstractMotif> allMotifs, 
			HierarchicalClusterer<AbstractMotif> clusterer,
			ConfigData config)
	{
		long start = System.currentTimeMillis();
		Logger.info("Clustering (threshold: %.2f)...\n", clusterer.threshold());
		ApplicationData appData = config.getAppData();
		String classOfExpectedJasparMotif = config.useJaspar() ? config.getJasparData().getExpectedClass() : null;
		
		clusterer.performClustering();
		List<HierarchicalCluster<AbstractMotif>> clusters = clusterer.getClusters();
		Collections.sort(clusters, new WeighableClustersComparator<AbstractMotif>());
		Collections.reverse(clusters);
		
		boolean success = false;
		int clustersWithFoundMotifs = 0;
		
		Logger.info("***************\n*** RESULTS ***\n***************\n");
		for (int i = 0; i < clusters.size(); i++){
			HierarchicalCluster<AbstractMotif> c = clusters.get(i);
			try {
				int findersCount = ClusterTools.getFindersNames(c).size();
				if (findersCount > 0){	
					/* This cluster contains at least one non-jaspar motif */
					clustersWithFoundMotifs ++;
					
					double avgClusterInfo = ClusterTools.getAverageInfo(c);
					ColumnsDistributionComparator cdc = getCDC(appData.getConsensusFunction()); 
					double [][] clusterPSPM = AlignmentTools.getBestAlignment(
							c.getObjects(), background, cdc);
					double icThreshold = appData.getIcThreshold(); 
					double [][] clusterCorePSPM = (icThreshold > 0) ?
							AlignmentTools.getCore(clusterPSPM, background, icThreshold):
							clusterPSPM;
					
					String clusterConsensus = QuickBioTools.getConsensusSequence(clusterCorePSPM);
					
					Logger.info("\nCLUSTER nr %2d (size: %d   avg info: %.2f   finders: %d)\n%s\n", 
						clustersWithFoundMotifs, c.size(), avgClusterInfo, findersCount, c);
					Logger.info("CONSENSUS: %s\n", clusterConsensus);
					if (appData.logPspmConsensus()){
						String pspmString = BioTools.toString(clusterCorePSPM);
						Logger.info("%s", pspmString);
					}
					Integer alignmentSize = appData.getAlignmentSize(); 
					if (alignmentSize != null && alignmentSize != 0){
						String clusterAlignment = AlignmentTools.getAlignmentAsWeblogoInput(clusterCorePSPM, alignmentSize);
						Logger.info("Data for WebLogo:\n%s\n", clusterAlignment);
					}
					
					if (classOfExpectedJasparMotif != null && ClusterTools.containsJasparMotif(c, classOfExpectedJasparMotif)){
						success = true;
						reportSuccess(c, i, findersCount, classOfExpectedJasparMotif);
					}
				}
			}  catch (WrongParameterException e) {
				throw new MMFRuntimeException(e);
			}
		}
		if (classOfExpectedJasparMotif != null && !success)
			Logger.auto("%s not found\n", classOfExpectedJasparMotif);

		HierarchicalCluster<AbstractMotif> c = clusterer.clusterToTheEnd();
		Logger.info("\nMAIN CLUSTER (weight: %.2f), NEWICK FORMAT:\n%s;\n", 
				ClusterTools.getTotalWeight(c), c.fullNewickFormat());
		Logger.info("\n%d ms total (clustering)\n", System.currentTimeMillis() - start);
	}


	
	private List<ExternalMotif> getExternalMotifs(ConfigData configData) throws IOException, BioException{

		String userMotifsFilePath = configData.getAppData().getUserMotifsFile();
		if (userMotifsFilePath == null){
			return new LinkedList<ExternalMotif>();
		}
		ExternalMotifsReader reader = new ExternalMotifsReader();
		return reader.readUserMotifs(userMotifsFilePath);
	}
	
	
	private List<ExternalMotif> getMotifsFromDatabase(ConfigData configData) throws IOException, BioException{

		String dbPath = configData.getAppData().getUserDatabase();
		if (dbPath == null){
			return new LinkedList<ExternalMotif>();
		}
		ExternalMotifsReader reader = new ExternalMotifsReader();
		return reader.readExternalDatabase(dbPath);
	}

	
	
	
	/**
	 * Get motifs from locally stored JASPAR database (location specified in config.xml, 
	 * in element <jaspar>). Only motifs from species specified in config.xml will be returned,
	 * and if nothing is specified there - program will wait for user's input. 
	 * 
	 * @param jasparData
	 * @return
	 * @throws IOException
	 * @throws BioException
	 */
	private List<JasparMotif> getJasparMotifs(ConfigData configData) throws IOException, BioException{

		if (!configData.useJaspar()){
			return new LinkedList<JasparMotif>();
		}

		JasparData jasparData = configData.getJasparData();
		ApplicationData appData = configData.getAppData();
		
		JasparDataReader jdr = new JasparDataReader(jasparData.getInfo(), 
				jasparData.getAnnotation(), jasparData.getData());
		Map<String, Collection<JasparMotif>> motifsMap = jdr.getMotifs(jasparData.isClassFiltering());

		if (configData.getAppData().getSpecies() != null){
			Collection<JasparMotif> jasparMotifs = motifsMap.get(appData.getSpecies().toLowerCase());
			if (jasparMotifs == null){
				Logger.warn("Species %s not found. No motifs from Jaspar added.\n", appData.getSpecies());
				return new LinkedList<JasparMotif>();
			} else {
				Logger.info("You've chosen species '%s' from Jaspar ==> clustering %d extra motifs\n\n", 
						appData.getSpecies(), jasparMotifs.size());
				return new LinkedList<JasparMotif>(jasparMotifs);
			}
		} else {

			Logger.info("The following organisms are available: \n");
			Map<Integer, Entry<String, Collection<JasparMotif>>> indexedMap = 
				new HashMap<Integer, Entry<String, Collection<JasparMotif>>>();
			int i = 1;
			int speciesCount = motifsMap.size();
			for (Entry<String, Collection<JasparMotif>> entry : motifsMap.entrySet()){
				Logger.info("  %2d. %s (%d motifs)\n", i, entry.getKey(), entry.getValue().size());
				indexedMap.put(i, entry);
				i++;
			}
			Logger.info("\n Please enter a number between %d and %d: ", 1, speciesCount);
			BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
			Integer index = null;
			while (index == null){
				try {
					index = new Integer(br.readLine());
					if (index < 1 || index > speciesCount){
						index = null;
					}
				} catch (NumberFormatException e){
					System.out.print("\nTry again: ");
				}
			}
			Entry<String, Collection<JasparMotif>> chosenEntry = indexedMap.get(index);
			Logger.info("You choose species %s ==> clustering %d extra motifs\n\n", 
					chosenEntry.getKey(), chosenEntry.getValue().size());
			return new LinkedList<JasparMotif>(chosenEntry.getValue());
		}
	}



	/**
	 * Runs each of the given motif finders to find motifs. After this operation,
	 * the results are printed to the stdout.
	 * 
	 * @param finders
	 * @param paramMap
	 * @param input
	 * @throws BioException 
	 * @throws IOException 
	 * @throws WrongParameterException 
	 */
	private List<Motif> findMotifs (
			ConfigData config,
			Map<String, RichSequence> seqMap,
			Sequence concatSeq,
			String sequenceFile, 
			String alphabetType,
			SimilarityFilter<AbstractMotif> filter) 
			throws BioException, IOException, MMFException, WrongParameterException
	{
		File sf = new File(sequenceFile);
		int lastDotIndex = sf.getAbsolutePath().lastIndexOf('.');
		String tmpPath = (lastDotIndex == -1) ? sf.getAbsolutePath() : sf.getAbsolutePath().substring(0, lastDotIndex);
		tmpFastaFile = BioTools.prepareSimplifiedFastaFile(tmpPath, ".sfa", seqMap);
		
		register(new WeederFinder());
		register(new MEMEFinder());
		register(new MDscanFinder());
		register(new BioProspectorFinder());

		/* set parameters for all finders */
		setParams(config, config.getFindersData().getParamMap());

		List<Motif> allMotifs = new LinkedList<Motif>();

		/* run only the finders registered in the parameters file */
		for (MotifFinderEntry mfe: finders.values()){
			if (!mfe.isRegistered())
				continue;
			MotifFinder f = mfe.getFinder();
			try {
				Logger.info("****************************************\n");
				long start = System.currentTimeMillis();
				List<Motif> motifs = f.findMotifs(tmpFastaFile, seqMap);
				long end = System.currentTimeMillis();
				Logger.info("%s found %d motifs (time: %d s).%n", f.getClass().getSimpleName(), motifs.size(), (end-start)/1000);

				int i = 0;
				for (Motif m: motifs){
					i++;
					m.prepareFittingProbabilities(concatSeq);
					Logger.info("Motif nr %d\n%s   avg info: %.4f bits/pos   occs: %d\n", 
							i, m.toString(), m.avgInfo(), m.getOccurrences().size());
					if (config.getAppData().logPspmMotifs())
						Logger.info(BioTools.toString(m.getDistributions()));
					
					/* Write all occurrences to the output */
					for (MotifOccurrence o: m.getOccurrences()){
						char direction = o.isForward() ? '+' : '-';
						SymbolList list = o.getSequence().subList(o.getPosition(), o.getPosition() + m.length() - 1);
						if (!o.isForward())
							list = DNATools.reverseComplement(list);
						Logger.debug("%s@%c%d %s  ", o.getSequence().getName(),direction, o.getPosition(), list.seqString());

					}
					Logger.debug("\n");
				}

				if (filter != null){
					filter.filter(motifs);
					Logger.info("\n%s: %d motifs left after refinement.\n", f.getClass().getSimpleName(), motifs.size());
				}

				allMotifs.addAll(motifs);
				Logger.info("****************************************\n");
			} catch (WrongValueReturnedException e){
				Logger.err("\nProgram %s failed with return value %d. No results from this program.\n\n", 
						e.getProgram(), e.getRetVal());
			} catch (MMFException e) {
				Logger.err("Motif finder %s failed because of the following exception:", f.getProgramIdentifier());
				e.printStackTrace();
			} catch (FileFormatException e){
				Logger.err("File: %s, line: %d, problem: %s", e.getFileName(), e.getLineNumber(), e.getMessage());
				e.printStackTrace();
			} catch (MMFRuntimeException er){
				Logger.err("Motif finder %s failed because of the following runtime exception:", f.getProgramIdentifier());
				er.printStackTrace();				
			}
		}
		Logger.info("Found %d motifs total.\n", allMotifs.size());
		return allMotifs;		
	}

	
	
	/**
	 * Removes the temporary simplified FASTA file. This is not required, but should be done 
	 * not to increase disk usage..
	 * 	 */
	private void clean(){
		if (tmpFastaFile == null) return;
		if (tmpFastaFile.exists()){
			boolean deleted = tmpFastaFile.delete();
			if (!deleted) 
				Logger.warn("Failed to delete file: %s.\n", tmpFastaFile.getAbsolutePath());
			else 
				Logger.info("Temporary file: %s deleted.\n", tmpFastaFile.getAbsolutePath());
		}
	}
	
	
	
	private ColumnsDistributionComparator getCDC(DISTRIBUTION_DISTANCE_FUNCTION function) throws WrongParameterException{
		switch (function){
		case DPQ:		return new QuickBioTools.CDpqCounter();
		case RELENT:	return new QuickBioTools.CRECounter();
		case PEARSON:	return  new QuickBioTools.CPCCounter();
		default:
			throw new WrongParameterException("Unsupported function (" + function + ") for comparing columns distributions");
		}
	}
	
	
	/**
	 * Returns FittingOnSequenceComparator specific for the given function
	 * @param function
	 * @return
	 * @throws WrongParameterException
	 */
	private FittingOnSequenceComparator getFoSC(DISTRIBUTION_DISTANCE_FUNCTION function) throws WrongParameterException{
		switch (function){
		case DPQ:		return new QuickBioTools.PDpqCounter(); 
		case RELENT:	return new QuickBioTools.PRECounter(); 
		case PEARSON:	return new QuickBioTools.PPCCounter(); 
		default:
			throw new WrongParameterException("Unsupported function (" + function + ") for comparing motifs' fitting distributions");
		}
	}
	
	

	private DistanceCounter<AbstractMotif> getDistanceCounter(
			DISTRIBUTION_DISTANCE_FUNCTION function,
			DISTRIBUTION_TYPE type,
			Map<String, RichSequence> seqMap,
			double gcContent) throws WrongParameterException
	{
		switch (type){
		
			case COLUMNS:
			{	
				final ColumnsDistributionComparator comparator = getCDC(function);
				double a = (1-gcContent)/2.0, t=a;
				double g = (gcContent) / 2.0, c=g;
				final Distribution bg = BioTools.getDNADistribution(a, c, g, t);
				
				return new DistanceCounter<AbstractMotif>(){
					public double distance(AbstractMotif m1, AbstractMotif m2) {
						return QuickBioTools.getColumnsDistributionDistance(m1, m2, comparator, bg);
					}
				};
			}
			case FITTING:
			{
				final FittingOnSequenceComparator comparator = getFoSC(function);
				
				return new DistanceCounter<AbstractMotif>(){
					public double distance(AbstractMotif m1, AbstractMotif m2) {
						try {
							return QuickBioTools.getFittingOnSequenceDistance(m1, m2, comparator);
						} catch (Exception e) {
							throw new MMFRuntimeException(e);
						}
					}
				};
			}
			default:
				return null;	
		}
	}


	
	/*
	 * Compares motifs each with each, and returns the distance between the closest two
	 */
	private double compareMotifsEachWithEach(List<? extends AbstractMotif> motifs,DistanceCounter<AbstractMotif> motifDistanceCounter){
		List<Pair<Pair<AbstractMotif, AbstractMotif>, Double>> l = QuickBioTools.sort(
				motifs, motifDistanceCounter);
		Logger.info("\n");
		for (Pair<Pair<AbstractMotif, AbstractMotif>, Double> p: l){
			Logger.info("%.6f  %s  %s\n", p.second(), p.first().first(), p.first().second());
		}
		return l.get(0).second();
	}
	
	
	
	private void run(String seqFile){

		try {
			ConfigData configData = ConfigSAXParser.parse(Constants.CONFIG_PATH);
			ApplicationData appData = configData.getAppData();
			Logger.setLogLevel(appData.getLogLevel());
			Constants.initialize(configData);
			background = QuickBioTools.getDistribution(appData.getGcContent());
			
			Logger.info("Length: %d   Reported: %d\n", 
					configData.getFindersData().getMotifLength(),
					configData.getFindersData().getMotifsReported());
			Logger.info("Metric: %s   Type: %s   Threshold: %.2f (%s)   Filter: %s\n", 
					appData.getFunction().toString(), 
					appData.getType(), 
					appData.getClusteringThresholdValue(), appData.getClusteringThresholdType(),
					configData.useJaspar() ? configData.getJasparData().isClassFiltering() : "<no jaspar>");
			
			Map<String, RichSequence> seqMap = BioTools.readFastaSequences(seqFile, "DNA");
			/* concatenate all sequences into one long */
			Sequence concatSeq = BioTools.concatenate( new LinkedList<RichSequence>(seqMap.values()), 
					"Concatenated:"+seqMap.size(), true);
			
			/* Initialize motifs from all sources */
			List<JasparMotif> jasparMotifs = getJasparMotifs(configData);
			for (JasparMotif m: jasparMotifs)
				m.prepareFittingProbabilities(concatSeq);
			List<ExternalMotif> externalMotifs = getExternalMotifs(configData);
			for (ExternalMotif m: externalMotifs)
				m.prepareFittingProbabilities(concatSeq);
			List<ExternalMotif> databaseMotifs = getMotifsFromDatabase(configData);
			for (ExternalMotif m: databaseMotifs)
				m.prepareFittingProbabilities(concatSeq);
			
			DistanceCounter<AbstractMotif> motifDistanceCounter = getDistanceCounter(
					appData.getFunction(), appData.getType(), seqMap, appData.getGcContent());

			if (configData.getFindersData().getMotifsReported() == 0){
				Logger.info("Zero motifs requested - comparing only jaspar motifs\n");
				compareMotifsEachWithEach(jasparMotifs, motifDistanceCounter);
				return;
			}
			
			SimilarityFilter<AbstractMotif> filter = null;
			Double refinementThreshold = appData.getRefinementThreshold();
			if (refinementThreshold != null)
				filter = new SimilarityFilter<AbstractMotif>(motifDistanceCounter, refinementThreshold);

			List<Motif> foundMotifs = findMotifs(configData, seqMap, concatSeq, seqFile, "DNA", filter);
			
			List<AbstractMotif> allMotifs = new ArrayList<AbstractMotif>();
			allMotifs.addAll(foundMotifs);
			Logger.info("Adding %d jaspar motifs...\n", jasparMotifs.size());
			allMotifs.addAll(jasparMotifs);
			Logger.info("Adding %d external motifs...\n", externalMotifs.size());
			allMotifs.addAll(externalMotifs);
			Logger.info("Adding %d motifs from user's database...\n", databaseMotifs.size());
			allMotifs.addAll(databaseMotifs);

			List<List<Double>> listsOfDistances = ClusterTools.countDistances(allMotifs, motifDistanceCounter);
			double clusteringThreshold;
			
			if (appData.getClusteringThresholdType() == THRESHOLD_TYPE.absolute)
				clusteringThreshold = appData.getClusteringThresholdValue();
			else {
				double maxDist;
				if (jasparMotifs.size() > 1){
					maxDist = QuickBioTools.getMaxDistanceBetweenSpecifiedMotifs(
							allMotifs, listsOfDistances, JasparMotif.IDENTIFIER);
				} else if (databaseMotifs.size() > 1){
					maxDist = QuickBioTools.getMaxDistanceBetweenSpecifiedMotifs(
							allMotifs, listsOfDistances, ExternalMotifsReader.EXTERNAL_DATABASE_MOTIFS);
				} else 
					throw new WrongParameterException("Too few (<= 1) motifs in database to count relative clustering threshold");
				clusteringThreshold = appData.getClusteringThresholdValue() * maxDist;
			}

			/* Perform clustering */
			switch(appData.getClusteringType()){
			case HIERARCHICAL:
				HierarchicalClusterer<AbstractMotif> clusterer = 
					new HierarchicalClusterer<AbstractMotif>(allMotifs,	listsOfDistances,
							clusteringThreshold);
				if (allMotifs.size() > 1)
					doClustering(allMotifs, clusterer, configData);
				break;
			case KMEDOIDS:
				throw new WrongParameterException("K-medoids clustering not yet supported.");
			}
		} catch (FileFormatException e) {
			Logger.err("Wrong format of file %s, line %d %s\n", e.getFileName(), e.getLineNumber(), 
					e.getMessage() == null ? "" : "(" + e.getMessage() + ")");
			Logger.err("Stacktrace:\n");
			e.printStackTrace();
			System.exit(-1);
		} catch (SAXParseException e){	// message has been already printed
			System.exit(-1);
		} catch (FileNotFoundException e){
			Logger.err("File not found: %s\n", e.getMessage());
			//Logger.err("Stacktrace:\n");
			//e.printStackTrace();
			System.exit(-1);
		} catch (Exception e) {
			e.printStackTrace();
			System.exit(-1);
		} finally {
			clean();
		}
	}


	
	public static void main(String[] args) {
		if (args.length == 0 || ((args.length == 1) && "-h".equals(args[0]))){
			help();
			System.exit(0);
		}
		long start = System.currentTimeMillis();
		new MMF().run(args[0]);
		long end = System.currentTimeMillis();
		Logger.info("\nEntire time: %d ms\n", end - start);
	}
	

	
	/**
	 * Print general help how to use this tool
	 *
	 */
	public static void help(){
		System.out.println("Usage: java MMF <fasta_file>");
		System.out.println("\tfasta_file:\tfile containing the sequences (in FASTA format)");
		System.exit(0);
	}
}