package mimuw.mmf.clusterers.kmedoids;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;

import mimuw.mmf.WrongParameterException;
import mimuw.mmf.clusterers.DistanceCounter;
import mimuw.mmf.finders.MMFRuntimeException;
import mimuw.mmf.util.Logger;
import mimuw.mmf.util.Tools;

public class KMedoidsClusterer <T> {

	private final int N;
	private final int K;
	private final List<KMedoidsCluster<T>> clusters;
	private final List<ObjectHolder<T>> objectHolders;
	private final int FAILED_ATTEMPTS_BEFORE_END = 8;
	private final double[][] distances;

	
	
	public KMedoidsClusterer (int clustersCount, List<T> objects, DistanceCounter<T> distanceCounter)
	{
		N = objects.size();
		K = clustersCount;
		clusters = new ArrayList<KMedoidsCluster<T>>(K);
		distances = new double[N][N];
		objectHolders = new ArrayList<ObjectHolder<T>>(N);
		for (int i = 0; i < N; i++)
			for (int j = 0; j < N; j++)
				distances[i][j] = distanceCounter.distance(objects.get(i), objects.get(j));

		for (int i = 0; i < N; i++){
			objectHolders.add(new ObjectHolder<T>(objects.get(i), i, false, null));
		}

		try {
			List<ObjectHolder<T>> medoids = Tools.sublist(K, objectHolders);
			for (ObjectHolder<T> medoidHolder: medoids){
				medoidHolder.isMedoid = true;
				KMedoidsCluster<T> newCluster = new KMedoidsCluster<T>(medoidHolder); 
				medoidHolder.cluster = newCluster;
				clusters.add(newCluster);
			}
			reassign();
			printClusters();
		} catch (WrongParameterException e) {
			throw new MMFRuntimeException(e);
		}
	}



	private void reassign(){
		//System.out.println("Number of clusters: " + clusters.size() );
		for (KMedoidsCluster<T> c : clusters){
			c.getObjects().clear();
		}

		List<ObjectHolder<T>> menoids = getMenoidsFromClusters(clusters);
		
		for (ObjectHolder<T> h : objectHolders){
			if (h.isMedoid){
				h.getCluster().getObjects().add(h);
			} else {
				KMedoidsCluster<T> bestCluster = findClosestMenoid(menoids, h).getCluster();
				h.setCluster(bestCluster);
				bestCluster.getObjects().add(h);
			}
		}
	}
	
	
	//TODO : warning it's expansive
	private double totalCosts(KMedoidsCluster<T> cluster){
		ObjectHolder<T> medoid = cluster.getMedoidHolder();
		double sum = 0;
		for (ObjectHolder<T> o : cluster.getObjects()){
			if (o != medoid){
				sum += distances[o.getIndex()][medoid.getIndex()];
			}
		}
		return sum;
	}
	
	
	private double totalCost(List<KMedoidsCluster<T>> clusters){
		double sum = 0;
		for (KMedoidsCluster<T> c : clusters){
			sum += totalCosts(c);
		}
		return sum;
	}
	
	
	
	private List<ObjectHolder<T>> getMenoidsFromClusters(List<KMedoidsCluster<T>> clusters){
		List<ObjectHolder<T>> menoids = new LinkedList<ObjectHolder<T>>();
		for (KMedoidsCluster<T> c: clusters){
			menoids.add(c.getMedoidHolder());
		}
		return menoids;
	}
	

	
	private double getAvgDistanceToOtherClusters(ObjectHolder<T> menoid, List<KMedoidsCluster<T>> clusters)
	{
		double sum = 0;
		int count = 0;
		double d[] = distances[menoid.getIndex()];
		for (KMedoidsCluster<T> c: clusters){
			if (c == menoid.getCluster()) continue;
			for (ObjectHolder<T> o: c.getObjects()){
				count ++;
				sum += d[o.getIndex()];
			}
		}
		return sum/count;
	}
	
	
	
	public double getAvgDistanceToOtherClusters(List<KMedoidsCluster<T>> clusters)
	{
		double sum = 0;
		int count = 0;

		for (KMedoidsCluster<T> c: clusters){
			double current = getAvgDistanceToOtherClusters(c.getMedoidHolder(), clusters);
			Logger.info("Avg distance from %s to the rest: %.1f\n", c.getMedoidHolder().getObject(), current);
			sum += current;
			count ++;
		}
		return sum/count;
	}
	
	
	

	private ObjectHolder<T> findClosestMenoid (List<ObjectHolder<T>> menoids, ObjectHolder<T> holder){

		double distMin = Double.MAX_VALUE;
		ObjectHolder<T> closestMenoid = null;
		double d[] = distances[holder.index];
		
		for (ObjectHolder<T> m: menoids){
			double dist = d[m.getIndex()];
			//System.out.printf("|%s-%s|=%.1f    ", holder.getObject(), m.getObject(), dist);
			if (dist < distMin){
				distMin = dist;
				closestMenoid = m;
			}
		}
		//System.out.printf("%s is closest to %s\n", holder.getObject(), closestMenoid.getObject());
		return closestMenoid;
	}



	private ObjectHolder<T> getRandomNonMedoid(List<ObjectHolder<T>> holders){
		ObjectHolder<T> result = null;
		Random rand = new Random();
		while (result == null || result.isMedoid()){
			result = holders.get(rand.nextInt(holders.size()));
		}
		return result;
	}



	private double countCostOfSwapping(
			List<KMedoidsCluster<T>> clusters, 
			ObjectHolder<T> newMedoid)
	{
		double total = 0;
		ObjectHolder<T> oldMenoid = newMedoid.getCluster().getMedoidHolder(); 
		
		List<ObjectHolder<T>> menoids = getMenoidsFromClusters(clusters);
		menoids.add(newMedoid);
		menoids.remove(oldMenoid);
		
		Logger.info("TRYING TO MAKE %s A MENOID\n", newMedoid.getObject());
		for (KMedoidsCluster<T> c: clusters){
			if (c != newMedoid.getCluster()){
				for (ObjectHolder<T> o: c.getObjects()){
					if (o.isMedoid()) continue;
					double d[] = distances[o.getIndex()];
					double oldDist = d[o.getCluster().getMedoidHolder().getIndex()];
					double newDist = d[newMedoid.getIndex()];
					double diff = newDist - oldDist;
					Logger.info("Diff for %s is: %.1f\n", o.getObject(), diff);
					if (diff < 0) // chance for improvement
						total += diff;
				}
			} else { // this cluster
				for (ObjectHolder<T> o: c.getObjects()){
					if (o == newMedoid || o == oldMenoid) continue;
					double d[] = distances[o.getIndex()];
					double oldDist = d[o.getCluster().getMedoidHolder().getIndex()];
					ObjectHolder<T> closestMenoid = findClosestMenoid(menoids, o);
					double newDist = d[closestMenoid.getIndex()];
					double diff = newDist - oldDist;
					System.out.printf("Diff for %s is: %.1f\n", o.getObject(), diff);
					total += diff;
				}
			}
		}
		return total;
	}


	public void performClustering(){
		int failed = 0;
		while (failed < FAILED_ATTEMPTS_BEFORE_END){
			ObjectHolder<T> candidateForMedoid = getRandomNonMedoid(objectHolders);
			double cost = countCostOfSwapping(clusters, candidateForMedoid);
			Logger.info("Cost for swapping %s with its medoid %s: %.2f\n", 
					candidateForMedoid.getObject(), candidateForMedoid.getCluster().getMedoidHolder().getObject(), cost);
			if (cost < 0){
				//System.out.printf("performClustering()::%s better than %s, cost: %f\n", candidateForMedoid.getObject(), candidateForMedoid.getCluster().getMedoidHolder().getObject(), cost);
				failed = 0;
				setAsMedoid(candidateForMedoid);
				printClusters();
			} else 
				//System.out.println("performClustering()::Medoid worse: " + cost);
			failed ++;
		}
		printClusters();
	}


	private void setAsMedoid(ObjectHolder<T> newMedoid){
		newMedoid.getCluster().getMedoidHolder().setMedoid(false);
		newMedoid.setMedoid(true);
		newMedoid.getCluster().setMedoidHolder(newMedoid);
		this.reassign();
	}



	private void printClusters(){
		Logger.info("\nCURRENT CLUSTERS (%.1f): \n", totalCost(clusters));
		//System.out.printf("\nCURRENT CLUSTERS: \n");
		for (KMedoidsCluster<T> cluster: clusters){
			String desc = cluster.toString();
			Logger.info(desc);
		}
		double avg = getAvgDistanceToOtherClusters(clusters);
		Logger.info("AVG DIST: %.1f\n", avg);
	}
	
	
	
	public List<KMedoidsCluster<T>> getClusters() {
		return clusters;
	}
}
