package mimuw.mmf.finders;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.biojavax.bio.seq.RichSequence;

import mimuw.mmf.motif.Motif;
import mimuw.mmf.util.Logger;

public abstract class AbstractResultsParser implements ResultsParserI {

	protected String fileName;
	protected BufferedReader bufRead;
	protected int count;
	protected String line = null;
	protected int motifIndex = 0;

	/**
	 * The only constructor. Tries to open file of the given name to read from.
	 * 
	 * @param resultsFile
	 * @throws MMFException
	 */
	protected AbstractResultsParser (String resultsFile) throws MMFException{
		count = 0;
		try {
			bufRead = new BufferedReader(new FileReader(resultsFile));
			fileName = resultsFile;
		} catch (FileNotFoundException e) {
			throw new MMFException("Temporary file "+ resultsFile + " not found.", e);
		} 
	}



	/**
	 * Read next line from the file that is being parsed. Returns this line, sets the attribute 'line'
	 * to the same string, and increases 'count'.
	 * 
	 * @return
	 */
	protected String readln() {
		try {
			line = bufRead.readLine();
		} catch (IOException e) {
			throw new MMFRuntimeException("I/O problem while reading line from " + fileName, e);
		}
		count++;
		return line;
	}



	protected abstract Motif parseMotif(Map<String, RichSequence> seqMap) throws MMFException;


	
	/**
	 * Parses the opened file and returns list of motifs. This method assumes that the format of the 
	 * file enables to read motifs one after another, and when no more motifs can be read - method
	 * readMotif() returns null. If this assumption is wrong for particular format, please override
	 * this method.
	 */
	public List<Motif> parseFile(Map<String, RichSequence> seqMap) throws MMFException{

		Logger.info("Parsing file: %s\n", fileName);
		List<Motif> motifsList = new ArrayList<Motif>();
		Motif motif;
		readln();
		while ((motif = parseMotif(seqMap)) != null){
			motifsList.add(motif);				
		}
		//BioTools.addWeigthsToMotifs(motifsList);
		clean();
		return motifsList;
	}


	protected void wrongFormat() throws FileFormatException{
		throw new FileFormatException(this.count, fileName);
	}

	protected void wrongFormat(String expected) throws FileFormatException{
		throw new FileFormatException("Expected: " + expected, this.count, fileName);
	}

	protected void wrongFormat(String expected, String found) throws FileFormatException {
		throw new FileFormatException("Expected: " + expected + ", found: " + found, this.count, fileName);
	}


	private void clean() {
		if (bufRead != null){
			try {
				bufRead.close();
			} catch (IOException e) {
				throw new MMFRuntimeException("I/O problem while closing file " + fileName, e);
			}
		}
	}
}
