package mimuw.mmf.finders.bioprospector;

import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Map;

import mimuw.mmf.WrongParameterException;
import mimuw.mmf.finders.AbstractMotifFinder;
import mimuw.mmf.finders.AbstractResultsParser;
import mimuw.mmf.finders.MMFException;
import mimuw.mmf.finders.WrongValueReturnedException;
import mimuw.mmf.motif.Motif;
import mimuw.mmf.util.Constants;
import mimuw.mmf.util.Logger;
import mimuw.mmf.util.Tools;

import org.biojavax.bio.seq.RichSequence;

public class BioProspectorFinder extends AbstractMotifFinder {

	public static final String BIOPROSPECTOR_PATH = Constants.FINDERS_PATH + "BioProspector/BioProspector.linux";
	public static final String IDENTIFIER = "bioprospector";
	
	public BioProspectorFinder() throws BioProspectorException {
		super();
	}



	public String getProgramIdentifier() {
		return IDENTIFIER;
	}


	
	 /** TODO: remove results file? */
	protected void clean(){}
	

	
	@Override
	public List<Motif> findMotifs(File seqFile, Map<String, RichSequence> seqMap) 
		throws WrongValueReturnedException, MMFException 
	{		
		List<Motif> result = null;
		StringBuffer cmd = new StringBuffer(100);
		Runtime rt = Runtime.getRuntime();
		String tmpResultsFile = Constants.TMP_RESULTS_DIR + seqFile.getName()+ ".results";
		Process proc;

		cmd .append(" " + BIOPROSPECTOR_PATH)
		.append(" -i " + seqFile.getAbsolutePath())
		.append(" -o " + tmpResultsFile)
		.append(getParametersString());

		try {
			Logger.info("BioProspectorFinder.findMotifs() - command: \n\t%s\n", cmd);
			proc = rt.exec(cmd.toString());
			int exitVal = Tools.executeProcessAndFlushStreams(proc);
			if (exitVal != 0)
				throw new WrongValueReturnedException(exitVal, getProgramIdentifier());
			AbstractResultsParser parser = new BioProspectorResultsParser(tmpResultsFile);
			result = parser.parseFile(seqMap);
		} catch (IOException e) {
			throw new BioProspectorException("Exception occured when looking for motifs with BioProspector",e);
		} catch (InterruptedException e) {
			throw new BioProspectorException("Exception occured when looking for motifs with BioProspector",e);
		} finally {
			clean();
		}

		return result;
	}	



	@Override
	public <T> void setParameter(T name, String val) throws WrongParameterException {
	
		if (name.toString().length() != 1)
			throw new WrongParameterException("BioProspector: Parameter name of length 1 required for each parameter (found " + name + ")");

		switch(name.toString().charAt(0)){
		case 'W':
			checkParameterRange(name, val, MIN_MOTIF_LEN, MAX_MOTIF_LEN);
			break;
		case 'o':
		case 'f':
		case 'b':
		case 'w':
			throw new WrongParameterException("Parameter -" + name + " is forbidden");
		case 'n':
		case 'r':
			checkParameterRange(name, val, 1, 200);
			break;
		case 'e':
			checkParameterRange(name, val, 1, 200); // TODO: '-e' should be limited by '-n' value 
			break;
		case 'p':
		case 'a':
			checkParameterRange(name, val, 0, 1);
			break;
		case 'T':
			checkParameterRange(name, val, MIN_REPORTED_MOTIFS_NUMBER, MAX_REPORTED_MOTIFS_NUMBER);
			break;
		case 'G':
		case 'g':
			checkParameterRange(name, val, 0, 50);
			break;
		case 'd':
			checkParameterRange(name, val, 1, 2);
			break;
		case 'h':
			checkParameterRange(name, val, 1, 1);
			break;
		case 'm':
		case 's':
			checkParameterPositiveFloat(name, val);
			break;
		case 'c':
			checkParameterRange(name, val, 1, 128);
			break;
		default:
			throw new WrongParameterException("Parameter -" + name + " is unknown");
		}

		putParameter(name, val);
	}
	

	@Override
	public void setParametersAsSingleString(String pS) throws WrongParameterException {
		String[] tokens = pS.split("-");
		for (String token: tokens){
			token = token.trim();
			if (token.equals(""))
				continue;
			String keyVal[] = token.split("\\s", 2);
			String name = keyVal[0];
			if (name.length() > 1)
				throw new WrongParameterException("BioProspector parameter's name cannot be longer than 1 (" + name + ")");
			String val = keyVal.length == 2 ? keyVal[1] : null;
			setParameter(name, val);
		}
	}


	public void setMotifLength(int len) throws WrongParameterException {
		setParameter("W", "" + len);
	}


	public void setSearchBothStrands(boolean both) {
			if (both)
				putParameter("d", "2");
			else
				putParameter("d", "1");
	}


	public void setDesiredNumberOfMotifs(int n) throws WrongParameterException {
		setParameter('T', "" + n);
	}
}
