package mimuw.mmf.jaspar;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import mimuw.mmf.finders.FileFormatException;
import mimuw.mmf.motif.TemporaryDatabaseMotif;
import mimuw.mmf.util.Logger;

import org.biojava.bio.BioException;
import org.biojava.bio.seq.DNATools;
import org.biojava.bio.symbol.AtomicSymbol;
import org.biojava.bio.symbol.IllegalSymbolException;
import org.biojava.utils.ChangeVetoException;

/**
 * JasparDataReader is responsible for reading JASPAR database from files into memory,
 * and building its internal representation
 * @author milosz
 *
 */
public class JasparDataReader {

	private final String infoFile;
	private final String annotationFile;
	private final String dataFile;

	private class TemporaryJasparMotif extends TemporaryDatabaseMotif {

		private static final long serialVersionUID = -3452805088498594579L;

		private String name;
		private String theClass;
		private String species;

		private TemporaryJasparMotif(String id) {
			super(id);
		}

		private void setSpecies(String species) {
			this.species = species;
		}
		private void setTheClass(String theClass) {
			this.theClass = theClass;
		}
		private void setName(String name) {
			this.name = name;
		}
		private JasparMotif toJasparMotif() throws BioException{
			return new JasparMotif(getId(), name, theClass, species, getTmpCounts());
		}
	}



	public JasparDataReader(final String infoFile, final String annotationFile, final String dataFile) {
		super();
		this.infoFile = infoFile;
		this.annotationFile = annotationFile;
		this.dataFile = dataFile;
	}



	/**
	 * Reads JASPAR information file (usually called MATRIX_INFO.txt). Each line should contain two entries 
	 * (separated by tabulator), first is motif's identifier (e.g. MA0095).
	 *  
	 * @param filename - absolute path to the info file
	 * @return
	 * @throws IOException
	 */
	private Map<String, TemporaryJasparMotif> readInfoFile(String filename) throws IOException{

		BufferedReader bufRead = new BufferedReader(new FileReader(filename));
		Map<String, TemporaryJasparMotif> result = new HashMap<String, TemporaryJasparMotif>();
		String line, tokens[], id;
		int lineNumber = 0;

		try {
			while ((line = bufRead.readLine()) != null){
				lineNumber ++;
				if (line.trim().equals(""))
					continue;
				tokens = line.split("\\s");
				if (tokens.length != 2)
					throw new FileFormatException(lineNumber, filename);
				id = tokens[0];
				result.put(id, new TemporaryJasparMotif(id));
			}
		} finally {
			bufRead.close();
		}

		return result;
	}



	/**
	 * Reads JASPAR annotation file (usually called MATRIX_ANNOTATION.txt). Each line contains
	 * motif's identifier, feature's name and feature's value, e.g.:
	 * MA0003	species	Homo sapiens
	 * or:
	 * MA0007	class	NUCLEAR RECEPTOR
	 * 
	 * @param filename - absolute path to the annotation file
	 * @param motifs - temporary motifs, already created by parsing INFO file 
	 * @throws IOException
	 */
	private void readAnnotationFile(String filename, Map<String, TemporaryJasparMotif> motifs) 
	throws IOException
	{
		BufferedReader bufRead = new BufferedReader(new FileReader(filename));
		String line, tokens[], id;
		int lineNumber = 0;

		try {
			while ((line = bufRead.readLine()) != null){
				lineNumber ++;
				if (line.trim().equals(""))
					continue;
				tokens = line.split("\\s", 3);
				if (tokens.length != 3)
					throw new FileFormatException(lineNumber, filename);
				id = tokens[0];
				if (tokens[1].equals("species"))
					motifs.get(id).setSpecies(tokens[2]);
				else if (tokens[1].equals("name"))
					motifs.get(id).setName(tokens[2]);
				else if (tokens[1].equals("class"))
					motifs.get(id).setTheClass(tokens[2]);
			}
		} finally {
			bufRead.close();
		}
	}

	

	/**
	 * Reads file describing all motifs in details(usually called MATRIX_DATA.txt). Each line consists of
	 * motif's identifier, nucleotide, position in motif (starting from 1) and occurrences count,
	 * for example:
	 * MA0065	A	5	17.0000
	 * Lines may appear in arbitrary order in this file.
	 * 
	 * @param filename - absolute path to the data file
	 * @param motifs - temporary motifs, already created by parsing INFO and ANNOTATION files
	 * @throws ChangeVetoException
	 * @throws IOException
	 */
	private <M extends TemporaryDatabaseMotif>void readDataFile(String filename, Map<String, M> motifs) 
	throws ChangeVetoException, IOException
	{
		BufferedReader bufRead = new BufferedReader(new FileReader(filename));
		String line, tokens[], id;
		AtomicSymbol symbol;
		int pos, lineNumber = 0;
		double count;
		M m;

		try {
			while ((line = bufRead.readLine()) != null){
				lineNumber ++;
				if (line.trim().equals(""))
					continue;
				tokens = line.split("\\s");
				if (tokens.length != 4)
					throw new FileFormatException(lineNumber, filename);
				id = tokens[0];
				m = motifs.get(id);
				if (m == null)
					throw new FileFormatException("Motif of given id not found", lineNumber, filename);
				if (tokens[1].length() != 1)
					throw new FileFormatException(lineNumber, filename);
				try {
					symbol = (AtomicSymbol) DNATools.forSymbol(tokens[1].charAt(0));
					pos = Integer.parseInt(tokens[2]);
					count = Double.parseDouble(tokens[3]);
					m.addData(symbol, pos, count);
				} catch (IllegalSymbolException e) {
					throw new FileFormatException(lineNumber, filename);
				} catch (NumberFormatException e) {
					throw new FileFormatException(lineNumber, filename);
				}
			}
		} finally {
			bufRead.close();
		}
	}
	
	

	private Map<String, Collection<JasparMotif>> convertIntoMapWithSpeciesAsKey(List<JasparMotif> motifs){

		final Map<String,  Collection<JasparMotif>> result = new HashMap<String,  Collection<JasparMotif>>();

		for (JasparMotif motif: motifs){
			Collection<JasparMotif> speciesStructure = result.get(motif.getSpecies().toLowerCase());
			if (speciesStructure == null){
				speciesStructure = new LinkedList<JasparMotif>();
				result.put(motif.getSpecies().toLowerCase(), speciesStructure);
			}
			speciesStructure.add(motif);
		}
		return result;
	}
	
	
	
	/**
	 * This method introduces some ordering into jaspar motifs. It gets a list of all motifs (from all species, all classes), 
	 * and transforms it into a map, where motif's species (lowercase) is a key, and a value is another map,
	 * where motif's class is a key, and best motif from this species and this class is a value. As a best motif we consider 
	 * such a motif, that has the highest average information on each position.
	 * @param motifs
	 * @return
	 */
	private Map<String, Map<String, JasparMotif>> toMapFromSpeciesToMapFromClassToBestMotif(List<JasparMotif> motifs){

		final Map<String, Map<String, JasparMotif>> result = new HashMap<String, Map<String,JasparMotif>>();

		for (JasparMotif motif: motifs){
			Map<String, JasparMotif> speciesStructure = result.get(motif.getSpecies().toLowerCase());
			if (speciesStructure == null){
				speciesStructure = new HashMap<String, JasparMotif>();
				speciesStructure.put(motif.getMotifsClass(), motif);
				result.put(motif.getSpecies().toLowerCase(), speciesStructure);
			} else {
				JasparMotif bestMotifOfThisClass = speciesStructure.get(motif.getMotifsClass());
				if (bestMotifOfThisClass == null || motif.avgInfo() > bestMotifOfThisClass.avgInfo())
					speciesStructure.put(motif.getMotifsClass(), motif);
			}
		}

		return result;
	}



	/**
	 * Flattens the complex structure returned by method toMapFromSpeciesToMapFromClassToBestMotif() to a simpler map,
	 * where species (lowercase) is a key, and collection of motifs belonging to this species is a value 
	 * (NOTE: there will be only one motif from each class)
	 * @param motifs
	 * @return
	 */
	private Map<String, Collection<JasparMotif>> flattenToMapFromSpeciesToMotifs (
			Map<String, Map<String, JasparMotif>> motifs)
		{

		Map<String, Collection<JasparMotif>> result = new HashMap<String, Collection<JasparMotif>>();

		for (Entry<String, Map<String, JasparMotif>> entry : motifs.entrySet()){
			String speciesLowercase = entry.getKey();
			Collection<JasparMotif> speciesSet = new LinkedList<JasparMotif>();
			//System.out.printf("Species: %s, motifs: %d\n", species, entry.getValue().values().size());
			for (JasparMotif m: entry.getValue().values()){
				speciesSet.add(m);
			}
			result.put(speciesLowercase, speciesSet);
		}

		return result;
	}



	/**
	 * Returns motifs from JASPAR database as a map where species (lowercase) is a key, 
	 * and collection of motifs belonging to this species is a value 
	 * (NOTE: there will be only one motif from each class - the one with the highest
	 * average information on position)
	 * @return
	 * @throws IOException
	 * @throws BioException
	 */
	public Map<String, Collection<JasparMotif>> getMotifs(boolean classFiltering) throws IOException, BioException{
		
		Map<String, TemporaryJasparMotif> motifsMap = readInfoFile(infoFile);
		
		readAnnotationFile(annotationFile, motifsMap);
		readDataFile(dataFile, motifsMap);
		List<JasparMotif> motifsList = new ArrayList<JasparMotif>(motifsMap.size());
		for (TemporaryJasparMotif m : motifsMap.values()){
			motifsList.add(m.toJasparMotif());
		}
		Logger.debug("Number of all motifs: %d\n", motifsList.size());
		
		Map<String, Collection<JasparMotif>> result;
		
		if (classFiltering){
			Map<String, Map<String, JasparMotif>> mapToRefine = toMapFromSpeciesToMapFromClassToBestMotif(motifsList);
			result = flattenToMapFromSpeciesToMotifs(mapToRefine);		
		} else {
			result = convertIntoMapWithSpeciesAsKey(motifsList);
		}
		return result;
	}
}