package mimuw.mmf.motif;

import java.util.Formatter;
import java.util.LinkedList;
import java.util.List;

import mimuw.mmf.clusterers.Weighable;
import mimuw.mmf.finders.MMFException;
import mimuw.mmf.finders.MMFRuntimeException;

import org.biojava.bio.dist.Count;
import org.biojava.bio.dist.IndexedCount;
import org.biojava.bio.seq.DNATools;
import org.biojava.bio.symbol.AtomicSymbol;
import org.biojava.bio.symbol.FiniteAlphabet;
import org.biojava.bio.symbol.IllegalSymbolException;
import org.biojava.utils.ChangeVetoException;
import org.biojavax.bio.seq.RichSequence;

/**
 * Motif is a class that represents motif found by a motif-finding program
 * 
 * @author milosz
 *
 */
public class Motif extends AbstractMotif implements Weighable {

	private final String finderId;
	private final List<MotifOccurrence> occurrences;
	private final int significanceIndex;	// index saying whether this motif was returned as 1st, second, etc
	
	
	private FiniteAlphabet alphabet;

	public Motif(int length, List<MotifOccurrence> occurrences, String finderId, int index) {
		super();
		this.length = length;
		this.finderId = finderId;
		if (occurrences.size() < 1){
			throw new MMFRuntimeException("No correct occurrences for this motif.");
		}
		this.occurrences = new LinkedList<MotifOccurrence>(occurrences);
		this.significanceIndex = index;
		init(toCounts(occurrences), mimuw.mmf.util.Constants.getBgDist());
	}

	public String getFinderId() {
		return finderId;
	}

	public List<MotifOccurrence> getOccurrences() {
		return occurrences;
	}

	public double getWeight() {
		return avgInfo();
	}

	@Deprecated
	public void setWeight(double weight) {
		throw new MMFRuntimeException("Method setWeight() unsupported");
		//this.weight = weight;
	}
	
	@Override
	public FiniteAlphabet getAlphabet() {
		return alphabet;
	}
	

	/**
	 * Counts the distribution of nucelotides/aminoacids/other symbols on each position 
	 * of the motif, and return the result as an array of distributions.
	 * 
	 * @return
	 * @throws NotInitializedException 
	 */
	private Count[] toCounts(List<MotifOccurrence> occurrences) {
		Count[] counts = new Count[length()];
		for (int i = 0; i < counts.length; i++){
			counts[i] = getCountAtPos(i);			
		}
		return counts;
	}
	

	
	/**
	 * Counts the distribution of nucelotides/aminoacids/other symbols on the specific position
	 * in motif. 
	 * 
	 * @param positionInMotif position between 0 and motif.length()-1
	 * @return
	 * @throws MMFException 
	 */
	private Count getCountAtPos (int positionInMotif) {

		// get the FiniteAlphabet from the first sequence
		FiniteAlphabet a = (FiniteAlphabet) getOccurrences().get(0).getSequence().getAlphabet();
		Count c = new IndexedCount(a);

		for (MotifOccurrence occ : getOccurrences()){
			RichSequence seq = occ.getSequence();
			int absolutePosition;
			AtomicSymbol symbol;
			if (occ.isForward()){
				absolutePosition = occ.getPosition() + positionInMotif;
				symbol = (AtomicSymbol) seq.symbolAt(absolutePosition);
			} else {
				absolutePosition = occ.getPosition() + length() - (positionInMotif +1);
				symbol = (AtomicSymbol) seq.symbolAt(absolutePosition);
				try {
					symbol = (AtomicSymbol) DNATools.complement(symbol);
				} catch (IllegalSymbolException e) {
					throw new MMFRuntimeException(e);
				}
			}

			try {
				c.increaseCount(symbol, 1.0);
			} catch (IllegalSymbolException e) {
				throw new MMFRuntimeException("Symbol " + symbol.getName() + " not found in Count made of alphabet " + a.getName(), e);
			} catch (ChangeVetoException e) {
				throw new MMFRuntimeException("Not allowed to change weight of symbol " + symbol.getName() + " in Count made of alphabet " + a.getName(), e);
			}
		}
		return c;
	}
	
	
	
	
	public String toStringVerbose(){
		StringBuffer buff = new StringBuffer();
		Formatter formatter = new Formatter(buff);
		formatter.format("Motif found by %s (index: %d), of length %d and weight %.3f", 
				getFinderId(), significanceIndex, length(), getWeight());
		return buff.toString();
	}


	@Override
	public String toString(){
		StringBuffer buff = new StringBuffer();
		Formatter formatter = new Formatter(buff);
		formatter.format("%s_%s_%d", consensus(), getFinderId().substring(0, 2), significanceIndex);
		return buff.toString();
	}

	public int getSignificanceIndex() {
		return significanceIndex;
	}

	@Override
	public String origin() {
		return this.getFinderId();
	}
}
