package mimuw.mmf.util;

import static mimuw.mmf.util.QuickBioTools.MINIMAL_OVERLAP;

import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import mimuw.mmf.finders.MMFRuntimeException;
import mimuw.mmf.motif.AbstractMotif;
import mimuw.mmf.util.QuickBioTools.ColumnsDistributionComparator;

public class AlignmentTools {

	public static double[][] getBestAlignment(
			List<? extends AbstractMotif> motifs, 
			double[] background,
			ColumnsDistributionComparator comp) 
	{
		Logger.trace(">>getBestAlignment() - aligning the following motifs:\n");
		sortMotifsByIC(motifs);
		double [][] result = motifs.get(0).getPSPM();
		int position = 1;
		for (AbstractMotif motif : motifs) {
			Logger.trace("%d. %s (%s) IC: %f \n", position, motif.toString(), motif.origin(), motif.avgInfo());
			Logger.trace("SAME STRAND:  ");
			Pair<Double, double[][]> p1 = getAlignmentWithBestDistance(
					motif.getPSPM(), result, background, 1, position, comp);
			Logger.trace("REVERSE COMP: ");
			Pair<Double, double[][]> p2 = getAlignmentWithBestDistance(
					motif.getRevCompPSPM(), result, background, 1, position, comp); 
				//getBestAlignment(motif.getRevCompPSPM(), result, background, 1, position, comp);
			result = (p1.first() < p2.first()) ? p1.second() : p2.second();
			position ++;
		}
		Logger.trace("<<getBestAlignment() - %d motifs aligned\n", position-1);
		return result;
	}
	

	@Deprecated
	protected static Pair<Double, double[][]> getAlignmentWithBestIC(
			double[][] pspm1,
			double[][] pspm2,
			double[] background, 
			double weight1,
			double weight2) {
		int len1 = pspm1.length;
		int len2 = pspm2.length;
		int maxLeftShift  = (len1 > MINIMAL_OVERLAP) ? (len1 - MINIMAL_OVERLAP) : 0;
		int maxRightShift = (len2 > MINIMAL_OVERLAP) ? (len2 - MINIMAL_OVERLAP) : 0;
		int bestShift = maxLeftShift;
		double bestIC = Double.MIN_VALUE;
		double bestAlignment[][] = null;
		for (int firstShiftedLeft = -maxRightShift; firstShiftedLeft <= maxLeftShift; firstShiftedLeft++){
			double [][] alignment = getConsensusPSPM(pspm1, pspm2, background, firstShiftedLeft, weight1, weight2);
			double ic = QuickBioTools.getAvgIC(alignment, background);
			Logger.debug("shift: %d    IC: %f\n", firstShiftedLeft, ic);
			if (ic > bestIC){
				bestIC = ic;
				bestAlignment = alignment;
				bestShift = firstShiftedLeft;
			}
		}
		Logger.debug("Best shift: %d\n", bestShift);
		return new Pair<Double, double[][]>(bestIC, bestAlignment);
	}
	

	protected static Pair<Double, double[][]> getAlignmentWithBestDistance(
			double[][] pspm1,
			double[][] pspm2,
			double[] background, 
			double weight1,
			double weight2,
			ColumnsDistributionComparator comp)
	{
		Pair<Integer,Double> p = QuickBioTools.getBestAvgColumnsDistributionDistance(comp, pspm1, pspm2, background);
		Logger.trace("Shift: %d   dist: %f\n", p.first(), p.second());
		double alignment[][] = getConsensusPSPM(pspm1, pspm2, background, -p.first(), weight1, weight2);
		return new Pair<Double, double[][]>(p.second(), alignment);
	}
	
	
	protected static double[][] getConsensusPSPM(
			double[][] motif1, 
			double[][] motif2, 
			double[] background, 
			int firstShiftedLeft, 
			double weight1, 
			double weight2)
	{
		
		if (firstShiftedLeft < 0){
			double tmp1[][] = motif2;
			motif2 = motif1;
			motif1 = tmp1;
			double tmp2 = weight2;
			weight2 = weight1;
			weight1 = tmp2;
			firstShiftedLeft *= -1;
		}
	
		int size, size1=motif1.length, size2=motif2.length;
		size = Math.max(firstShiftedLeft+size2, size1);
		
		double result [][] = new double[size][4];
		double first  [][] = new double[size][4];
		double second [][] = new double[size][4];
		
		int i = 0;
		for (; i < firstShiftedLeft; i++){
			first[i] = motif1[i];
			second[i] = background;
		}
		for (; i < size1 && i < size2+firstShiftedLeft; i++){
			first[i] = motif1[i];
			second[i] = motif2[i-firstShiftedLeft];
		}
		
		if (i == size1){
			for (; i < size; i++){
				first[i] = background;
				second[i] = motif2[i-firstShiftedLeft];
			}	
		} else {
			first[i] = motif1[i];
			second[i] = background;
		}
		
		for (int pos = 0; pos < size; pos++)
			result[pos] = getConsensus(first[pos], second[pos], weight1, weight2);
		
		return result;
	}
	
	
	protected static void sortMotifsByIC(List<? extends AbstractMotif> motifs){
		
		if (motifs.isEmpty()) return;
		Collections.sort(motifs, new Comparator<AbstractMotif>(){
			
			public int compare(AbstractMotif o1, AbstractMotif o2) {
				if (o1.avgInfo() < o2.avgInfo()) return -1;
				if (o1.avgInfo() > o2.avgInfo()) return 1;
				return 0;
			}	
		});
		Collections.reverse(motifs);
	}
	
	
	private static double[] getConsensus(double[] p1, double[] p2, double weight1, double weight2){
		double result[] = new double[4];
		double wsum = weight1 + weight2;
		for (int i = 0; i < 4; i++)
			result[i] = (weight1/wsum)*p1[i] + (weight2/wsum)*p2[i];
		return result;
	}


	public static double[][] getCore(double[][] pspm, double[] background, double icThreshold) {

		int start=0, end=pspm.length-1;
		while ((start < pspm.length) && (icThreshold > QuickBioTools.getIC(pspm[start], background))){
			start ++;
		}
		while (end >= 0 && (icThreshold > QuickBioTools.getIC(pspm[end], background))){
			end --;
		}
		if (start >= end){
			Logger.info("No positions with IC > %.2f in the consensus.\n", icThreshold);
			return pspm;
		}
		double [][] result = new double[end-start+1][4];
		System.arraycopy(pspm, start, result, 0, end-start+1);
		return result;
	}
	
	
	public static String getAlignmentAsWeblogoInput(double [][] pspm, int sequences){
		
		Logger.trace(">>getAlignmentAsWeblogoInput()\n");
		if (sequences < 1)
			throw new MMFRuntimeException("Can't produce alignment for " + sequences + " sequences");
		double [][] copy = pspm.clone();
		int seqLen = pspm.length;
		
		String representations[] = new String[seqLen];
		/* representations[i] - a textual representation of distribution at position i  */
		for (int pos = 0; pos < seqLen; pos++){
			representations[pos] = QuickBioTools.getRepresentation(copy[pos], sequences);
			Logger.trace("representations[%d]=%s\n", pos, representations[pos]);
		}
		StringBuffer buf = new StringBuffer();
		
		for (int pos = 0; pos < sequences; pos++){
			for (int seq = 0; seq < seqLen; seq ++){
				String column = representations[seq];
				buf.append(column.charAt(pos));
			}
			buf.append('\n');
		}
		Logger.trace("<<getAlignmentAsWeblogoInput()\n");
		return buf.toString();
	}
}
