package mimuw.mmf.util;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import mimuw.mmf.finders.MMFException;
import mimuw.mmf.motif.AbstractMotif;
import mimuw.mmf.motif.Motif;
import mimuw.mmf.motif.MotifOccurrence;

import org.biojava.bio.BioException;
import org.biojava.bio.seq.DNATools;
import org.biojava.bio.symbol.Alphabet;
import org.biojavax.bio.seq.RichSequence;
import org.junit.Test;

public class AlignmentToolsTest {

	@Test
	public void testGetConsensusPSPM(){
		double bg[] =     {0.27, 0.23, 0.27, 0.23};
		double motif1[][] = {{0.5, 0.1, 0.1, 0.3}, {0.1, 0.1, 0.1, 0.7}};
		double motif2[][] =                      {{0.01, 0.09, 0.1, 0.8}, {0.4, 0.3, 0.2, 0.1}};
		
		int firstShiftedLeft = 1;
		double weight1 = 3;
		double weight2 = 1;
		double wsum = weight1 + weight2;
		double alignment[][] = AlignmentTools.getConsensusPSPM(motif1, motif2, bg, 
				firstShiftedLeft, weight1, weight2);
		assertEquals(3, alignment.length);
		assertEquals(motif1[firstShiftedLeft][3]*weight1/wsum + motif2[0][3]*weight2/wsum, 
				alignment[1][3]);
		
		for (int i = 0; i < alignment.length; i++){
			System.out.printf("\n alignment[%d]:\t", i);
			for (int j = 0; j < 4; j++)
				System.out.printf("%.4f\t", alignment[i][j]);
		}
		System.out.println();
	}

	@Test
	public void testGetBestAlignment() throws BioException, MMFException{
		
		Alphabet dnaAlphabet = DNATools.getDNA();
		double[] bg = {0.3, 0.2, 0.2, 0.3};
		
		Map<String, RichSequence> seqMap = new HashMap<String, RichSequence>();
		RichSequence seqA = RichSequence.Tools.createRichSequence("seqA", "tAGTTCAcacc", dnaAlphabet);
		RichSequence seqB = RichSequence.Tools.createRichSequence("seqB", "gggAGTTCAccaa", dnaAlphabet);
		RichSequence seqC = RichSequence.Tools.createRichSequence("seqC", "tAGTTCAcaccgtgtgtc", dnaAlphabet);
		seqMap.put(seqA.getName(), seqA);
		seqMap.put(seqB.getName(), seqB);
		seqMap.put(seqC.getName(), seqC);

		List<MotifOccurrence> occurrences = new LinkedList<MotifOccurrence>();
		for (int i = 0; i < 2; i++){
			occurrences.add(new MotifOccurrence(seqA, 3, true, 6)); // taGTGCACacc 
			occurrences.add(new MotifOccurrence(seqB, 5, true, 6)); // gggaGTGCACaa
			occurrences.add(new MotifOccurrence(seqC, 3, true, 6)); // taGTGCACaccgtgtgtc
		}
		Motif m1 = new Motif(6, occurrences, "testingFinder", 1);
	
		occurrences = new LinkedList<MotifOccurrence>();
		for (int i = 0; i < 2; i++){
			occurrences.add(new MotifOccurrence(seqA, 2, true, 6)); // aAGTTCAcacc 
			occurrences.add(new MotifOccurrence(seqB, 4, true, 6)); // gggAGTTCAccaa
			occurrences.add(new MotifOccurrence(seqC, 4, true, 6)); // tagTTCACaccgtgtgtc
		}
		Motif m2 = new Motif(6, occurrences, "testingFinder", 2);	

		occurrences = new LinkedList<MotifOccurrence>();
		for (int i = 0; i < 2; i++){
			occurrences.add(new MotifOccurrence(seqA, 1, true, 5)); // TAGTT
			occurrences.add(new MotifOccurrence(seqC, 1, true, 5)); // TAGTT
		}
		Motif m3 = new Motif(5, occurrences, "testingFinder", 3);
			
		List<AbstractMotif> motifs = new ArrayList<AbstractMotif>();
		motifs.add(m1);
		motifs.add(m2);
		motifs.add(m3);
		double alignment[][] = AlignmentTools.getBestAlignment(motifs, bg, new QuickBioTools.CRECounter());
		System.out.printf("Results:\n%s\n", BioTools.toString(alignment));
	}

	@Test
	public void testGetCore(){
		double[] bg = {0.25, 0.25, 0.25, 0.25};
		double motif1[][] = {
				bg,
				{0.6, 0.1, 0.1, 0.2}, 
				{0.1, 0.1, 0.1, 0.7}, bg,
				{0.1, 0.1, 0.2, 0.6}, bg, bg
		};
		double expected[][] = {
				{0.6, 0.1, 0.1, 0.2}, 
				{0.1, 0.1, 0.1, 0.7}, bg,
				{0.1, 0.1, 0.2, 0.6},
		};
		double [][]result = AlignmentTools.getCore(motif1,bg,0.4);
		assertEquals("Wrong size of PSPM core", expected.length, result.length);
		for (int pos = 0; pos < result.length; pos++){
			for (int nucl = 0; nucl < 4; nucl++){
				assertEquals(expected[pos][nucl], result[pos][nucl]);
			}
		}
	}

	@Test
	public void testGetAlignment(){
		double pspm[][] = {
				{0, 1, 0, 0},
				{0.25, 0.25, 0.25, 0.25},
				{0.5, 0, 0.5, 0},
				{1, 0, 0, 0},
				{0, 0, 0, 1}
		};
		int size = 4;
		String alignment = AlignmentTools.getAlignmentAsWeblogoInput(pspm, size);
		String alignmentWords[] = alignment.split("\n");
		String aw[] = alignmentWords;
		assertEquals(4, alignmentWords.length);
		Arrays.sort(alignmentWords);
		
		assertTrue(aw[0].charAt(0) == 'c');
		assertTrue(aw[1].charAt(0) == 'c');
		assertTrue(aw[2].charAt(0) == 'c');
		assertTrue(aw[3].charAt(0) == 'c');
		
		int pos = 1;
		int a=0, c=0, g=0, t=0;
		for (int w = 0; w < 4; w++){
			switch (aw[w].charAt(pos)){
			case 'a': a++; break;
			case 'c': c++; break;
			case 'g': g++; break;
			case 't': t++; break;
			}
		}
		assertEquals(1, a);
		assertEquals(1, c);
		assertEquals(1, g);
		assertEquals(1, t);
		
		pos++;
		a=0; c=0; g=0; t=0;
		for (int w = 0; w < 4; w++){
			switch (aw[w].charAt(pos)){
			case 'a': a++; break;
			case 'c': c++; break;
			case 'g': g++; break;
			case 't': t++; break;
			}
		}
		assertEquals(2, a);
		assertEquals(0, c);
		assertEquals(2, g);
		assertEquals(0, t);
		
		assertTrue(aw[0].charAt(3) == 'a');
		assertTrue(aw[1].charAt(3) == 'a');
		assertTrue(aw[2].charAt(3) == 'a');
		assertTrue(aw[3].charAt(3) == 'a');
		
		assertTrue(aw[0].charAt(4) == 't');
		assertTrue(aw[1].charAt(4) == 't');
		assertTrue(aw[2].charAt(4) == 't');
		assertTrue(aw[3].charAt(4) == 't');
	}
}
