package mimuw.mmf.util;

import static org.junit.Assert.assertEquals;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;

import mimuw.mmf.WrongParameterException;
import mimuw.mmf.finders.MMFException;
import mimuw.mmf.motif.Motif;
import mimuw.mmf.motif.MotifOccurrence;
import org.biojava.bio.BioException;
import org.biojava.bio.seq.DNATools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.symbol.FiniteAlphabet;
import org.biojava.utils.ChangeVetoException;
import org.biojavax.bio.seq.RichSequence;
import org.junit.Ignore;
import org.junit.Test;


public class QuickBioToolsTest {

	private FiniteAlphabet dnaAlphabet = DNATools.getDNA();
	
	@Test
	public void testPearsonsCorrelation1(){
		double d1[] = {1,2,4,6.6};
		double d2[] = {3,6,12,19.8};
		double result = QuickBioTools.PearsonsCorrelation(d1, d2, 0);
		double expected = 1;
		assertEquals(expected, result, BioTools.EPSILON);
	}
	
	@Test
	public void testPearsonsCorrelation2(){
		double d1[] = {0.3, 0.2, 0.3};
		double d2[] = {0.24, 0.24, 0.24};
		double result = QuickBioTools.PearsonsCorrelation(d1, d2, 0);
		System.out.printf("CPC between almost uniform and uniform: %f\n", result);
	}
	
	@Test
	public void testPearsonsCorrelationMinus1(){
		double d1[] = {1,2,4,8};
		double d2[] = {-1,-2,-4,-8};
		double result = QuickBioTools.PearsonsCorrelation(d1, d2, 0);
		double expected = -1;
		assertEquals(expected, result, BioTools.EPSILON);
	}
	
	@Test
	public void testPearsonsCorrelationDiff(){
		double d1[] = {3.3, 8.1, 4.5, 0.7, 5.5};
		double d2[] = {9.2, 4.2, 3.9, 2.4, 6.8};
		QuickBioTools.PearsonsCorrelation(d1, d2, 0);
	}
	
	@Test
	public void testCompareMotifsOnSequences() throws BioException, MMFException, WrongParameterException{
		
		Map<String, RichSequence> seqMap = new HashMap<String, RichSequence>();
		RichSequence seqA = RichSequence.Tools.createRichSequence("seqA", "aAGTGCAcacc", dnaAlphabet);
		RichSequence seqB = RichSequence.Tools.createRichSequence("seqB", "gggAGTGCAccaa", dnaAlphabet);
		RichSequence seqC = RichSequence.Tools.createRichSequence("seqC", "tAGTGCAcaccgtgtgtc", dnaAlphabet);
		seqMap.put(seqA.getName(), seqA);
		seqMap.put(seqB.getName(), seqB);
		seqMap.put(seqC.getName(), seqC);

		List<MotifOccurrence> occurrences = new LinkedList<MotifOccurrence>();
		occurrences.add(new MotifOccurrence(seqA, 2, true, 6)); // aAGTGCAcacc 
		occurrences.add(new MotifOccurrence(seqB, 4, true, 6)); // gggAGTGCAccaa
		occurrences.add(new MotifOccurrence(seqC, 2, true, 6)); // tAGTGCAcaccgtgtgtc
		Motif m1 = new Motif(6, occurrences, "testingFinder", 1);
			
		occurrences = new LinkedList<MotifOccurrence>();
		occurrences.add(new MotifOccurrence(seqA, 2, true, 6)); // aAGTGCAcacc 
		occurrences.add(new MotifOccurrence(seqB, 4, true, 6)); // gggAGTGCAccaa
		occurrences.add(new MotifOccurrence(seqC, 4, true, 6)); // tagTGCACaccgtgtgtc
		Motif m2 = new Motif(6, occurrences, "testingFinder", 2);
		
		System.out.printf("Motif m1: \n%s", BioTools.toString(m1.getDistributions()));
		System.out.printf("Motif m2: \n%s", BioTools.toString(m2.getDistributions()));
		
		double pm[] = QuickBioTools.getMotifProbsOnSingleSequence(m1.getPSPM(), seqA);
		double pm2[] = QuickBioTools.getMotifProbsOnSingleSequence(m2.getPSPM(), seqA);
		
		System.out.printf("Scores of motif m1 on sequence %s:\n", seqA.getName());
		for (int i =0; i < pm.length; i++){
			System.out.printf("%d: %.8f   ", i, pm[i]);
		}
		System.out.println();
		
		System.out.printf("Scores of motif m2 on sequence %s:\n", seqA.getName());
		for (int i =0; i < pm2.length; i++){
			System.out.printf("%d: %.8f   ", i, pm2[i]);
		}
		System.out.println();
	}
	
	
	@Test
	public void testCountMotifsProbabilityAtPosition() throws BioException, ChangeVetoException, SecurityException, NoSuchMethodException, IllegalArgumentException, IllegalAccessException, InvocationTargetException{
		
		double d0[] =  {0.1, 0.1, 0.8, 0};
		double d1[] =  {0.05, 0.9, 0.03, 0.02};
		double d[][] = {d0, d1};
		
		Sequence seq = RichSequence.Tools.createRichSequence("sequenceTested", "ACGTGC", dnaAlphabet);
		int position = 5;
		
		// Below I use reflection to pass by the modifier 'private'
		Class<?> paramClasses [] = {double[][].class, Sequence.class, Integer.class};
		Method methodToTest = QuickBioTools.class.getDeclaredMethod("countMotifsProbabilityAtPosition", paramClasses);
		methodToTest.setAccessible(true);
		Object params [] = {d, seq, position};
		Double result = (Double) methodToTest.invoke(null, params);
		methodToTest.setAccessible(false);
		//double result = BioTools.countMotifsProbabilityAtPosition(d, seq, position);
		double expected = 0.8 * 0.9;
		assertEquals(expected, result);
	}
	
	@SuppressWarnings("deprecation")
	@Test
	public void testDpq() throws WrongParameterException{
		System.out.println("Testing Dpq");
		double d0[] =  {0.01, 0.59, 0.28, 0.12};
		double d1[] =  {0.25, 0.16, 0.00, 0.59};
		double relent = QuickBioTools.relativeEntropy(d0, d1,0);
		double result = QuickBioTools.DpqDistance(d0, d1);
		double result2 = QuickBioTools.quickDpqDistance(d0, d1, 0);
		double pearson = QuickBioTools.PearsonsCorrelationToDistance(
				QuickBioTools.PearsonsCorrelation(d1, d0, 0));
		System.out.printf("RelEnt: %f, Dpq: %f, %f, Pearson's: %f\n", 
				relent, result, result2, pearson);
		assertEquals(result, result2);
	}
	
	@Test
	public void testDpqSymmetry() throws WrongParameterException{
		System.out.println("Testing Dpq symmetry");
		double d0[] =  {0.01, 0.59, 0.28, 0.12};
		double d1[] =  {0.25, 0.16, 0.02, 0.57};
		double result1 = QuickBioTools.quickDpqDistance(d0, d1, 0);
		double result2 = QuickBioTools.quickDpqDistance(d1, d0, 0);
		assertEquals(result1, result2);
	}
	
	@Test @Ignore
	public void testGetMotifsProbs() throws BioException, WrongParameterException{
		double d0[] =  {1, 0, 0, 0};
		double d1[] =  {0, 0.6, 0.4, 0};
		double d2[] =  {0, 1, 0, 0};
		double d3[] =  {0.8, 0, 0, 0.2};
		double d4[] =  {0, 0, 1, 0};
		double d5[] =  {0, 0, 0, 1};
		double[][] motif = {d0, d1, d2, d3, d4, d5};
		
		double e0[] = {0.875, 0.042, 0.042, 0.042};
		double e1[] = {0.042, 0.542, 0.375, 0.042};
		double e2[] = {0.042, 0.875, 0.042, 0.042};
		double e3[] = {0.708, 0.042, 0.042, 0.208};
		double e4[] = {0.042, 0.042, 0.875, 0.042};
		double e5[] = {0.042, 0.042, 0.042, 0.875};
		
		double[][] motif2 = {e0, e1, e2, e3, e4, e5};
		
		RichSequence seqA = RichSequence.Tools.createRichSequence("seqA",
				"ggaccagccggtctacgtcatctctcatagaggattctaacTCCGGAAGatttggatg" +
				"ctacgtaattctcatgtttattgagacccgggaggtcggaggtctagatatgctaaat" +
				"actccacgcacgggattatacACACCTGCccgggggttttaaaatgtcgggtctctac" +
				"tgttctcaaccgcagttgatatgcaattacggcctccgacccaagctaagcatgctga" +
				"cgcgcactctaaaatgttgtgtcagttatcgcgcaACACCTGCacctgtctccaaatc" +
				"tgctcagtagaagctggtgACCAGTAGtgattcctcgaggtggactacgatgcaacat" +
				"actttgcgtctcatgccattaaacaggctcctatgccatggtctatcgcgattgtaa",
				//"CAGCCTGTGATCTTTGCCTCACCAGTTTGGGCTCCTCCACCGGCACGAGTTGTCAGATTGTGTTTCCGGCTGCCTCTTCCCGTTGACCCCCTCCTCCCCCAACACCTGTCCCCTCTCCCGCCCACCCTCATTCCACAGCCCTGTAGACAGG",
				//-*-----------------+-----------------*-----*----------------------------------------------------------+-----------------------------------------------
				dnaAlphabet);
		double probs[] = QuickBioTools.getMotifProbsOnSingleSequence(motif, seqA);
		int i = 0;
		for (double pr: probs){
			//if (pr > 0)
			i++;
			System.out.printf("%d %.4f\n", i, pr);
		}
		System.out.println("\n\n");
		
		i = 0;
		double probs2[] = QuickBioTools.getMotifProbsOnSingleSequence(motif2, seqA);
		for (double pr: probs2){
			i++;
			System.out.printf("%d %.6f\n", i, pr);
		}
		System.out.println();
	}
	
	@SuppressWarnings("deprecation")
	@Test
	public void testBothDpqs() throws WrongParameterException{
		int SIZE = 1000000;
		double [] motif1, motif2;
		Random rand = new Random();
		
		motif1 = new double[SIZE];
		motif2 = new double[SIZE];
		for (int i = 0; i < SIZE; i++){
			motif1[i] = rand.nextDouble();
			motif2[i] = rand.nextDouble();
		}
		
		long time0 = System.currentTimeMillis();
		double res1 = QuickBioTools.DpqDistance(motif1, motif2);
		long time1 = System.currentTimeMillis();
		double res2 = QuickBioTools.quickDpqDistance(motif1, motif2, 0);
		long time2 = System.currentTimeMillis();
		double resultPC = QuickBioTools.PearsonsCorrelationToDistance(QuickBioTools.PearsonsCorrelation(motif1, motif2,0));
		long time3 = System.currentTimeMillis();
		
		System.out.printf("DPQ time: %d, DPQ2: %d  PC: %d\n", time1-time0, time2-time1, time3-time2);
		System.out.printf("Results DPQ: %.2f, DPQ2: %.2f, PC: %.2f\n", res1, res2, resultPC);
		assertEquals(res1, res2);
	}
	
	
	@Test
	public void testNormalize() throws WrongParameterException{
		double d[] = {1,2,5,4,10,2,7,9,5,5}; //sum=50
		double result[] = QuickBioTools.normalize(d, 0, 9);
		assertEquals(result[0], 0.02);
		assertEquals(result[9], 0.1);
		result = QuickBioTools.normalize(d, 8, 9);
		assertEquals(result.length, 2);
		assertEquals(result[1], 0.5);
	}
	
	@Test
	public void testNormalize2() throws WrongParameterException{
		double d[] = {1,2,5,4,10,2,7,9,5,5}; //sum=50
		double result[] = QuickBioTools.normalize(d, 50.0, 0, 9);
		assertEquals(result[0], 0.02);
		assertEquals(result[9], 0.1);
		result = QuickBioTools.normalize(d, 50.0, 8, 9);
		assertEquals(result.length, 2);
		assertEquals(result[1], 0.5);
	}
	
	@Test
	public void testIC() {
		double bg [] = {0.3, 0.2, 0.2, 0.3};
		double ic = QuickBioTools.getIC(bg, bg);
		assertEquals(0, ic);
	}
	
	@Test
	public void testIC2() {
		double bg [] = {0.3, 0.2, 0.2, 0.3};
		double d []  = {0, 0.8, 0.2, 0};
		double ic = QuickBioTools.getIC(d, bg);
		System.out.printf("InfC: %f\n", ic);
		assertEquals(1.6, ic);
	}
	
	@Test
	public void testAvgIC() {
		double bg [] = {0.3, 0.2, 0.2, 0.3};
		double pspm [][] = {bg, {0, 0.8, 0.2, 0}};
		double ic = QuickBioTools.getAvgIC(pspm, bg);
		System.out.printf("IC: %f\n", ic);
		assertEquals(0.8, ic);
	}
	
	@Test
	public void testGetRepresentation(){
		double d[] = {0.3, 0.1, 0.5, 0.1};
		int size = 10;
		String expected = "aaacgggggt";
		String actual = QuickBioTools.getRepresentation(d, size);
		assertEquals(expected, actual);
		size = 5;
		expected = "acggt";
		actual = QuickBioTools.getRepresentation(d, size);
		assertEquals(expected, actual);
		size = 20;
		expected = "aaaaaaccggggggggggtt";
		actual = QuickBioTools.getRepresentation(d, size);
		assertEquals(expected, actual);
	}
}