package mimuw.mmf.finders.weeder;

import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Map;

import mimuw.mmf.WrongParameterException;
import mimuw.mmf.finders.MotifFinder;
import mimuw.mmf.finders.MMFException;
import mimuw.mmf.finders.MMFRuntimeException;
import mimuw.mmf.finders.WrongValueReturnedException;
import mimuw.mmf.motif.Motif;
import mimuw.mmf.util.Constants;
import mimuw.mmf.util.Logger;
import mimuw.mmf.util.Tools;

import org.biojavax.bio.seq.RichSequence;

public class WeederFinder implements MotifFinder {

	private static final String WEEDER_PATH = Constants.FINDERS_PATH + "Weeder1.3/";
	public static final String IDENTIFIER = "weeder";
	
	private enum ANALYSIS_TYPE {small, medium, large, extra};

	// required parameters
	private ANALYSIS_TYPE analType;
	private String speciesCode;

	// optional parameters
	private boolean allSequencesMustContain;
	private boolean processBothStrands;
	private boolean multipleInEachSeq;
	private int topMotifsReportedPerRun;


	public WeederFinder() {
		super();
		topMotifsReportedPerRun = 5;
		analType = ANALYSIS_TYPE.medium;
	}


	public String getProgramIdentifier() {
		return IDENTIFIER;
	}


	public List<Motif> findMotifs(File seqFile, Map<String, RichSequence> seqMap) 
		throws WrongValueReturnedException, MMFException 
	{

		Logger.trace(">>WeederFinder.findMotifs()\n");

		List<Motif> result;
		StringBuffer cmd = new StringBuffer(100);
		Runtime rt = Runtime.getRuntime();
		String weederResultsFile = seqFile + ".wee";
		String weederResultsHTMLFile = seqFile + ".html";
		
		// if file that results will be appended to already exists, delete it
		new File(weederResultsFile).delete();
		new File(weederResultsHTMLFile).delete();
		
		cmd .append(WEEDER_PATH + "weederlauncher.out ")
			.append(seqFile.getAbsolutePath())
			.append(getParametersString());
		
		try {
			Logger.info("WeederFinder.findMotifs() - command: \n\t%s\n", cmd);
			// Working dir is required as Weeder looks for the ./FreqFiles/
			File workingDir = new File(WEEDER_PATH);
			Logger.debug("Working dir: %s\n", workingDir.getAbsolutePath());
			Process proc = rt.exec(cmd.toString(), null, workingDir);
			int exitVal = Tools.executeProcessAndFlushStreams(proc);
			if (exitVal != 0)
				throw new WrongValueReturnedException(exitVal, getProgramIdentifier());
			WeederResultsParser parser = new WeederResultsParser(weederResultsFile);
			result = parser.parseFile(seqMap);
		} catch (IOException e) {
			throw new MMFException("IO exception", e);
		} catch (InterruptedException e) {
			throw new MMFException("Interrupted exception", e);
		} 

		Logger.trace("<<WeederFinder.findMotifs()\n");
		return result;
	}



	private String getParametersString() {
		
		StringBuffer buf = new StringBuffer();
		buf.append(' ');
		buf.append(speciesCode);
		buf.append(' ');
		buf.append(analType);
		if (allSequencesMustContain) buf.append(" A");
		if (processBothStrands) buf.append(" S");
		if (multipleInEachSeq) buf.append(" M");
		buf.append(" T" + topMotifsReportedPerRun);
		Logger.trace("paramString: %s\n", buf.toString());
		return buf.toString();
	}


	public void setParameter(String key, String val) {
		throw new MMFRuntimeException("Cannot set single parameter for weeder");
	}


	
	/**
	 * Sets all the parameters for Weeder as one string. The string must contain the species code
	 * (e.g. HS, MM or BBS), the analysis type (small|medium|large|extra), and then - any additional
	 * paramters (A, S, M, T <number>)
	 */
	public void setParametersAsSingleString(String pS) throws WrongParameterException {

		String tokens [] = pS.split("\\s");
		if (tokens.length < 2){
			throw new WrongParameterException("Parameters for weeder required (at least organism + analysis type)");
		}

		this.speciesCode = tokens[0].trim();

		try {
			int i = 1;
			if (tokens[1].startsWith("sm") || tokens[1].startsWith("me") || tokens[1].startsWith("la") || tokens[1].startsWith("ex")){
				this.analType = ANALYSIS_TYPE.valueOf(tokens[1].trim().toLowerCase());
				i=2;
			}
			for (; i < tokens.length; i++){
				String keyVal = tokens[i].trim();
				
				char param = keyVal.charAt(0);
				switch (param){
				case 'A':
					allSequencesMustContain = true;
					break;
				case 'S':
					processBothStrands = true;
					break;
				case 'M':
					multipleInEachSeq = true;
					break;
				case 'T':
					String val = keyVal.substring(1);
					if (!Tools.isIntegerBetween(val, MIN_REPORTED_MOTIFS_NUMBER, MAX_REPORTED_MOTIFS_NUMBER))
						throw new WrongParameterException("Wrong value of parameter T: " + val);
					topMotifsReportedPerRun = new Integer(val);
					break;
				default:
					throw new WrongParameterException("Unknown argument for weeder analysis type (" + keyVal + ")");
				}
			}
		} catch (IllegalArgumentException e){
			throw new WrongParameterException("Wrong value for weeder analysis type (" + tokens[1].trim() + ")");
		}
	}


	public void setMotifLength(int len) throws WrongParameterException {
		if (len < MIN_MOTIF_LEN || len > MAX_MOTIF_LEN)
			throw new WrongParameterException("Wrong length of motif ("+len + ")");
		if (len <= 9)
			analType = ANALYSIS_TYPE.small;
		else if (len <= 11)
			analType = ANALYSIS_TYPE.medium;
		else
			analType = ANALYSIS_TYPE.large;
	}


	public void setSearchBothStrands(boolean both) {
		processBothStrands = both;
	}


	public void setDesiredNumberOfMotifs(int n) throws WrongParameterException {
		if (n < MIN_REPORTED_MOTIFS_NUMBER || n > MAX_REPORTED_MOTIFS_NUMBER)
			throw new WrongParameterException("Wrong number of motifs: " + n);
		topMotifsReportedPerRun = n;
	}
}
